@@ -33,7 +33,7 @@
 #include "item_windowfunc.h"
 
 bool mysql_union(THD *thd, LEX *lex, select_result *result,
-                 SELECT_LEX_UNIT *unit, ulong setup_tables_done_option)
+                 SELECT_LEX_UNIT *unit, ulonglong setup_tables_done_option)
 {
   DBUG_ENTER("mysql_union");
   bool res;
@@ -2436,8 +2436,8 @@ bool st_select_lex_unit::exec()
       fake_select_lex->table_list.empty();
       if (likely(!saved_error))
       {
-	thd->limit_found_rows = (ulonglong)table->file->stats.records + add_rows;
-        thd->inc_examined_row_count(examined_rows);
+        thd->limit_found_rows=
+            (ulonglong) table->file->stats.records + add_rows;
       }
       /*
 	Mark for slow query log if any of the union parts didn't use
@@ -2448,6 +2448,8 @@ bool st_select_lex_unit::exec()
   thd->lex->current_select= lex_select_save;
 err:
   thd->lex->set_limit_rows_examined();
+  if (likely(!saved_error))
+    thd->inc_examined_row_count(examined_rows);
   DBUG_RETURN(saved_error);
 }
 
@@ -2622,7 +2624,8 @@ bool st_select_lex_unit::cleanup()
       With_element *with_elem= with_element;
       while ((with_elem= with_elem->get_next_mutually_recursive()) !=
              with_element)
-        with_elem->rec_result->cleanup_count++;
+        if (with_elem->rec_result)
+          with_elem->rec_result->cleanup_count++;
       DBUG_RETURN(FALSE);
     }
   }
