@@ -145,7 +145,8 @@ int main(int argc, char **argv)
   {
     if ((ma_control_file_open(FALSE, opt_require_control_file ||
                               !(check_param.testflag & T_SILENT),
-                              TRUE)))
+                              TRUE,
+                              control_file_open_flags)))
     {
       if (opt_require_control_file ||
           (opt_transaction_logging && (check_param.testflag & T_REP_ANY)))
@@ -235,9 +236,6 @@ end:
   }
   maria_end();
   my_exit(error);
-#ifndef _lint
-  return 0;				/* No compiler warning */
-#endif
 } /* main */
 
 enum options_mc {
@@ -506,7 +504,7 @@ static void usage(void)
 		      maria_chk very silent.\n\
   -t, --tmpdir=path   Path for temporary files. Multiple paths can be\n\
                       specified, separated by ");
-#if defined( __WIN__) || defined(__NETWARE__)
+#if defined( _WIN32)
    printf("semicolon (;)");
 #else
    printf("colon (:)");
@@ -987,7 +985,7 @@ static void get_options(register int *argc,register char ***argv)
 
   if (set_collation_name)
     if (!(set_collation= get_charset_by_name(set_collation_name,
-                                             MYF(MY_WME))))
+                                             MYF(MY_UTF8_IS_UTF8MB3 | MY_WME))))
       my_exit(1);
 
   if (maria_data_root != default_log_dir && opt_log_dir == default_log_dir)
@@ -1593,8 +1591,14 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
     }
     compile_time_assert((MY_UUID_STRING_LENGTH + 1) <= sizeof(buff));
     buff[MY_UUID_STRING_LENGTH]= 0;
-    my_uuid2str(share->base.uuid, buff);
+    my_uuid2str(share->base.uuid, buff, 1);
     printf("UUID:                %s\n", buff);
+    if (ma_control_file_inited() &&
+        memcmp(share->base.uuid, maria_uuid, MY_UUID_SIZE))
+      printf("Warning: File UUID not match control file UUID! "
+             "File is probably moved\n"
+             "It will be updated to new system on first usage if zerofill is "
+             "not done\n");
     pos=buff;
     if (share->state.changed & STATE_CRASHED)
       strmov(buff, share->state.changed & STATE_CRASHED_ON_REPAIR ?
